import Foundation
import TSCBasic
import TuistCore
import TuistGraph
import TuistGraphTesting
import XCTest

@testable import TuistCoreTesting
@testable import TuistGeneratorTesting
@testable import TuistKit
@testable import TuistLintingTesting
@testable import TuistLoaderTesting
@testable import TuistSupportTesting

final class LintCodeServiceTests: TuistUnitTestCase {
    private var codeLinter: MockCodeLinter!
    private var manifestGraphLoader: MockManifestGraphLoader!
    private var basePath: AbsolutePath!

    private var subject: LintCodeService!

    override func setUpWithError() throws {
        try super.setUpWithError()

        codeLinter = MockCodeLinter()
        manifestGraphLoader = MockManifestGraphLoader()

        basePath = try temporaryPath()

        subject = LintCodeService(
            codeLinter: codeLinter,
            manifestGraphLoader: manifestGraphLoader
        )
    }

    override func tearDown() {
        subject = nil

        codeLinter = nil
        manifestGraphLoader = nil
        basePath = nil

        super.tearDown()
    }

    func test_run_throws_an_error_when_target_no_exist() async throws {
        // Given
        let project = Project.test(path: basePath.appending(component: "test"))
        let target01 = Target.test(name: "Target1")
        let target02 = Target.test(name: "Target2")
        let target03 = Target.test(name: "Target3")
        let graph = Graph.test(
            projects: [project.path: project],
            targets: [
                project.path: [
                    target01.name: target01,
                    target02.name: target02,
                    target03.name: target03,
                ],
            ]
        )
        let fakeNoExistTargetName = "Target_999"
        manifestGraphLoader.stubLoadGraph = graph

        // When
        await XCTAssertThrowsSpecific(
            try await subject.run(path: project.path.pathString, targetName: fakeNoExistTargetName, strict: false),
            LintCodeServiceError.targetNotFound(fakeNoExistTargetName)
        )
    }

    func test_run_throws_an_error_when_target_to_lint_has_no_sources() async throws {
        // Given
        let project = Project.test(path: basePath.appending(component: "test"))
        let target01 = Target.test(name: "Target1", sources: [])
        let target02 = Target.test(name: "Target2", sources: [])
        let target03 = Target.test(name: "Target3", sources: [])
        let graph = Graph.test(
            projects: [project.path: project],
            targets: [
                project.path: [
                    target01.name: target01,
                    target02.name: target02,
                    target03.name: target03,
                ],
            ]
        )
        manifestGraphLoader.stubLoadGraph = graph

        // When
        await XCTAssertThrowsSpecific(
            try await subject.run(path: project.path.pathString, targetName: target01.name, strict: false),
            LintCodeServiceError.lintableFilesForTargetNotFound(target01.name)
        )
    }

    func test_run_throws_an_error_when_code_liner_throws_error() async throws {
        // Given
        let fakeError = TestError("codeLinterFailed")
        let project = Project.test(path: basePath.appending(component: "test"))
        codeLinter.stubbedLintError = fakeError

        // When
        await XCTAssertThrowsSpecific(
            try await subject.run(path: project.path.pathString, targetName: nil, strict: false),
            fakeError
        )
    }

    func test_run_lint_workspace() async throws {
        // Given
        let workspace = Workspace.test(path: basePath.appending(component: "test"))
        let project = Project.test(path: basePath.appending(component: "test"))
        let target01 = Target.test(
            name: "Target1",
            sources: [
                SourceFile(path: "/target01/file1.swift", compilerFlags: nil),
                SourceFile(path: "/target01/file2.swift", compilerFlags: nil),
            ]
        )
        let target02 = Target.test(
            name: "Target2",
            sources: [
                SourceFile(path: "/target02/file1.swift", compilerFlags: nil),
                SourceFile(path: "/target02/file2.swift", compilerFlags: nil),
                SourceFile(path: "/target02/file3.swift", compilerFlags: nil),
            ]
        )
        let target03 = Target.test(
            name: "Target3",
            sources: [
                SourceFile(path: "/target03/file1.swift", compilerFlags: nil),
            ]
        )
        let graph = Graph.test(
            workspace: workspace,
            projects: [project.path: project],
            targets: [
                project.path: [
                    target01.name: target01,
                    target02.name: target02,
                    target03.name: target03,
                ],
            ]
        )
        manifestGraphLoader.stubLoadGraph = graph

        // When
        try await subject.run(path: workspace.path.pathString, targetName: nil, strict: false)

        // Then
        let invokedLintParameters = codeLinter.invokedLintParameters

        XCTAssertEqual(codeLinter.invokedLintCount, 1)
        XCTAssertEqual(
            Set(invokedLintParameters?.sources ?? []),
            [
                "/target01/file1.swift",
                "/target01/file2.swift",
                "/target02/file1.swift",
                "/target02/file2.swift",
                "/target02/file3.swift",
                "/target03/file1.swift",
            ]
        )
        XCTAssertEqual(invokedLintParameters?.path, project.path)
        XCTAssertEqual(invokedLintParameters?.strict, false)

        XCTAssertPrinterOutputContains("""
        Loading the dependency graph at \(workspace.path)
        Running code linting
        """)
    }

    func test_run_lint_project() async throws {
        // Given
        let project = Project.test(path: basePath.appending(component: "test"))
        let target01 = Target.test(
            name: "Target1",
            sources: [
                SourceFile(path: "/target01/file1.swift", compilerFlags: nil),
                SourceFile(path: "/target01/file2.swift", compilerFlags: nil),
            ]
        )
        let target02 = Target.test(
            name: "Target2",
            sources: [
                SourceFile(path: "/target02/file1.swift", compilerFlags: nil),
                SourceFile(path: "/target02/file2.swift", compilerFlags: nil),
                SourceFile(path: "/target02/file3.swift", compilerFlags: nil),
            ]
        )
        let target03 = Target.test(
            name: "Target3",
            sources: [
                SourceFile(path: "/target03/file1.swift", compilerFlags: nil),
            ]
        )
        let graph = Graph.test(
            projects: [project.path: project],
            targets: [
                project.path: [
                    target01.name: target01,
                    target02.name: target02,
                    target03.name: target03,
                ],
            ]
        )
        manifestGraphLoader.stubLoadGraph = graph

        // When
        try await subject.run(path: project.path.pathString, targetName: nil, strict: false)

        // Then
        let invokedLintParameters = codeLinter.invokedLintParameters

        XCTAssertEqual(codeLinter.invokedLintCount, 1)
        XCTAssertEqual(
            Set(invokedLintParameters?.sources ?? []),
            [
                "/target01/file1.swift",
                "/target01/file2.swift",
                "/target02/file1.swift",
                "/target02/file2.swift",
                "/target02/file3.swift",
                "/target03/file1.swift",
            ]
        )
        XCTAssertEqual(invokedLintParameters?.path, project.path)
        XCTAssertEqual(invokedLintParameters?.strict, false)

        XCTAssertPrinterOutputContains("""
        Loading the dependency graph at \(project.path)
        Running code linting
        """)
    }

    func test_run_lint_project_strict() async throws {
        // Given
        let project = Project.test(path: basePath.appending(component: "test"))
        let target01 = Target.test(
            name: "Target1",
            sources: [
                SourceFile(path: "/target01/file1.swift", compilerFlags: nil),
                SourceFile(path: "/target01/file2.swift", compilerFlags: nil),
            ]
        )
        let target02 = Target.test(
            name: "Target2",
            sources: [
                SourceFile(path: "/target02/file1.swift", compilerFlags: nil),
                SourceFile(path: "/target02/file2.swift", compilerFlags: nil),
                SourceFile(path: "/target02/file3.swift", compilerFlags: nil),
            ]
        )
        let target03 = Target.test(
            name: "Target3",
            sources: [
                SourceFile(path: "/target03/file1.swift", compilerFlags: nil),
            ]
        )
        let graph = Graph.test(
            projects: [project.path: project],
            targets: [
                project.path: [
                    target01.name: target01,
                    target02.name: target02,
                    target03.name: target03,
                ],
            ]
        )
        manifestGraphLoader.stubLoadGraph = graph

        // When
        try await subject.run(path: project.path.pathString, targetName: nil, strict: true)

        // Then
        let invokedLintParameters = codeLinter.invokedLintParameters

        XCTAssertEqual(codeLinter.invokedLintCount, 1)
        XCTAssertEqual(
            Set(invokedLintParameters?.sources ?? []),
            [
                "/target01/file1.swift",
                "/target01/file2.swift",
                "/target02/file1.swift",
                "/target02/file2.swift",
                "/target02/file3.swift",
                "/target03/file1.swift",
            ]
        )
        XCTAssertEqual(invokedLintParameters?.path, project.path)
        XCTAssertEqual(invokedLintParameters?.strict, true)

        XCTAssertPrinterOutputContains("""
        Loading the dependency graph at \(project.path)
        Running code linting
        """)
    }

    func test_run_lint_target() async throws {
        // Given
        let workspace = Workspace.test(path: basePath.appending(component: "test"))
        let project = Project.test(path: basePath.appending(component: "test"))
        let target01 = Target.test(
            name: "Target1",
            sources: [
                SourceFile(path: "/target01/file1.swift", compilerFlags: nil),
                SourceFile(path: "/target01/file2.swift", compilerFlags: nil),
            ]
        )
        let target02 = Target.test(
            name: "Target2",
            sources: [
                SourceFile(path: "/target02/file1.swift", compilerFlags: nil),
                SourceFile(path: "/target02/file2.swift", compilerFlags: nil),
                SourceFile(path: "/target02/file3.swift", compilerFlags: nil),
            ]
        )
        let target03 = Target.test(
            name: "Target3",
            sources: [
                SourceFile(path: "/target03/file1.swift", compilerFlags: nil),
            ]
        )
        let graph = Graph.test(
            workspace: workspace,
            projects: [project.path: project],
            targets: [
                project.path: [
                    target01.name: target01,
                    target02.name: target02,
                    target03.name: target03,
                ],
            ]
        )
        manifestGraphLoader.stubLoadGraph = graph

        // When
        try await subject.run(path: workspace.path.pathString, targetName: target01.name, strict: false)

        // Then
        let invokedLintParameters = codeLinter.invokedLintParameters

        XCTAssertEqual(codeLinter.invokedLintCount, 1)
        XCTAssertEqual(invokedLintParameters?.sources, [
            "/target01/file1.swift",
            "/target01/file2.swift",
        ])
        XCTAssertEqual(invokedLintParameters?.path, workspace.path)
        XCTAssertEqual(invokedLintParameters?.strict, false)

        XCTAssertPrinterOutputContains("""
        Loading the dependency graph at \(workspace.path)
        Running code linting
        """)
    }
}
